from fntypes.result import Error, Ok, Result

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter.abc import ABCAdapter
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.model import Model
from telegrinder.types.objects import Update


class RawEventAdapter(ABCAdapter[Update, Model]):
    def __init__(self, event_model: type[Model], /) -> None:
        self.event_model = event_model

    def __repr__(self) -> str:
        return "<{}: adapt Update -> {}>".format(
            self.__class__.__name__,
            self.event_model.__name__,
        )

    def adapt(self, api: API, update: Update, context: Context) -> Result[Model, AdapterError]:
        if isinstance(update.incoming_update, self.event_model):
            return Ok(update.incoming_update)
        return Error(AdapterError(f"Update is not an {self.event_model.__name__!r}."))


__all__ = ("RawEventAdapter",)



import typing

from telegrinder.node.base import ComposeError, FactoryNode, ScalarNode
from telegrinder.node.message import MessageNode


class Text(ScalarNode, str):
    @classmethod
    def compose(cls, message: MessageNode) -> str:
        if not message.text:
            raise ComposeError("Message has no text.")
        return message.text.unwrap()


class TextInteger(ScalarNode, int):
    @classmethod
    def compose(cls, text: Text) -> int:
        if not text.isdigit():
            raise ComposeError("Text is not digit.")
        return int(text)


if typing.TYPE_CHECKING:
    from typing import Literal as TextLiteral

else:

    class TextLiteral(FactoryNode):
        texts: tuple[str, ...]

        def __class_getitem__(cls, texts, /):
            return cls(texts=(texts,) if not isinstance(texts, tuple) else texts)

        @classmethod
        def compose(cls, text: Text) -> str:
            if text in cls.texts:
                return text
            raise ComposeError("Text matching failed.")


__all__ = ("Text", "TextInteger", "TextLiteral")



import abc
import typing

from telegrinder.types.objects import Message as MessageEvent

from .abc import ABCRule, CheckResult, Message
from .adapter import EventAdapter


class MessageRule(ABCRule[Message], abc.ABC):
    adapter: EventAdapter[Message] = EventAdapter(MessageEvent, Message)

    @abc.abstractmethod
    def check(self, *args: typing.Any, **kwargs: typing.Any) -> CheckResult: ...


__all__ = ("MessageRule",)



import base64
import dataclasses
import enum
import keyword
import os
import secrets
import typing
from datetime import datetime
from types import NoneType

import msgspec
from fntypes.co import Nothing, Result, Some

from telegrinder.msgspec_utils import decoder, encoder, get_origin

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError

T = typing.TypeVar("T")
P = typing.ParamSpec("P")

UnionType: typing.TypeAlias = typing.Annotated[tuple[T, ...], ...]

MODEL_CONFIG: typing.Final[dict[str, typing.Any]] = {
    "omit_defaults": True,
    "dict": True,
    "rename": {kw + "_": kw for kw in keyword.kwlist},
}


@typing.overload
def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: type[T],
) -> Result[T, "APIError"]: ...


@typing.overload
def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: UnionType[T],
) -> Result[T, "APIError"]: ...


def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: typing.Any,
) -> Result[typing.Any, "APIError"]:
    return result.map(lambda v: decoder.decode(v, type=full_t))


def generate_random_id(length_bytes: int) -> str:
    if length_bytes < 1 or length_bytes > 64:
        raise ValueError("Length of bytes must be between 1 and 64.")

    random_bytes = os.urandom(length_bytes)
    random_id = base64.urlsafe_b64encode(random_bytes).rstrip(b"=").decode("utf-8")
    return random_id


def get_params(params: dict[str, typing.Any]) -> dict[str, typing.Any]:
    validated_params = {}
    for k, v in (
        *params.pop("other", {}).items(),
        *params.items(),
    ):
        if isinstance(v, Proxy):
            v = v.get()
        if k == "self" or type(v) in (NoneType, Nothing):
            continue
        validated_params[k] = v.unwrap() if isinstance(v, Some) else v
    return validated_params


if typing.TYPE_CHECKING:

    @typing.overload
    def field(*, name: str | None = ...) -> typing.Any: ...

    @typing.overload
    def field(*, default: typing.Any, name: str | None = ...) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default_factory: typing.Callable[[], typing.Any],
        name: str | None = None,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = ...,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default: typing.Any,
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = ...,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default_factory: typing.Callable[[], typing.Any],
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = None,
    ) -> typing.Any: ...

    def field(
        *,
        default=...,
        default_factory=...,
        name=...,
        converter=...,
    ) -> typing.Any: ...

    class From(typing.Generic[T]):
        def __new__(cls, _: T, /) -> typing.Any: ...
else:
    from msgspec import field as _field

    From = typing.Annotated[T, ...]

    def field(**kwargs):
        kwargs.pop("converter", None)
        return _field(**kwargs)


@typing.dataclass_transform(field_specifiers=(field,))
class Model(msgspec.Struct, **MODEL_CONFIG):
    @classmethod
    def from_data(cls: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        return decoder.convert(msgspec.structs.asdict(cls(*args, **kwargs)), type=cls)  # type: ignore

    @classmethod
    def from_dict(cls, obj: dict[str, typing.Any], /) -> typing.Self:
        return decoder.convert(obj, type=cls)

    @classmethod
    def from_bytes(cls, obj: bytes, /) -> typing.Self:
        return decoder.decode(obj, type=cls)

    def _to_dict(
        self,
        dct_name: str,
        exclude_fields: set[str],
        full: bool,
    ) -> dict[str, typing.Any]:
        if dct_name not in self.__dict__:
            self.__dict__[dct_name] = (
                msgspec.structs.asdict(self)
                if not full
                else encoder.to_builtins(self.to_dict(exclude_fields=exclude_fields), order="deterministic")
            )

        if not exclude_fields:
            return self.__dict__[dct_name]

        return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        """
        :param exclude_fields: Model field names to exclude from the dictionary representation of this model.
        :return: A dictionary representation of this model.
        """

        return self._to_dict("model_as_dict", exclude_fields or set(), full=False)

    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        """
        :param exclude_fields: Model field names to exclude from the dictionary representation of this model.
        :return: A dictionary representation of this model including all models, structs, custom types.
        """

        return self._to_dict("model_as_full_dict", exclude_fields or set(), full=True)


@dataclasses.dataclass(kw_only=True, frozen=True, slots=True, repr=False)
class DataConverter:
    _converters: dict[type[typing.Any], typing.Callable[..., typing.Any]] = dataclasses.field(
        init=False,
        default_factory=lambda: {},
    )
    _files: dict[str, tuple[str, bytes]] = dataclasses.field(default_factory=lambda: {})

    def __repr__(self) -> str:
        return "<{}: {}>".format(
            self.__class__.__name__,
            ", ".join(f"{k}={v.__name__!r}" for k, v in self._converters.items()),
        )

    def __post_init__(self) -> None:
        self._converters.update(
            {
                get_origin(value.__annotations__["data"]): value
                for key, value in vars(self.__class__).items()
                if key.startswith("convert_") and callable(value)
            }
        )

    def __call__(self, data: typing.Any, *, serialize: bool = True) -> typing.Any:
        converter = self.get_converter(get_origin(type(data)))
        if converter is not None:
            if isinstance(converter, staticmethod):
                return converter(data, serialize)
            return converter(self, data, serialize)
        return data

    @property
    def converters(self) -> dict[type[typing.Any], typing.Callable[..., typing.Any]]:
        return self._converters.copy()

    @property
    def files(self) -> dict[str, tuple[str, bytes]]:
        return self._files.copy()

    @staticmethod
    def convert_enum(data: enum.Enum, _: bool = False) -> typing.Any:
        return data.value

    @staticmethod
    def convert_datetime(data: datetime, _: bool = False) -> int:
        return int(data.timestamp())

    def get_converter(self, t: type[typing.Any]):
        for type_, converter in self._converters.items():
            if issubclass(t, type_):
                return converter
        return None

    def convert_model(
        self,
        data: Model,
        serialize: bool = True,
    ) -> str | dict[str, typing.Any]:
        converted_dct = self(data.to_dict(), serialize=False)
        return encoder.encode(converted_dct) if serialize is True else converted_dct

    def convert_dct(
        self,
        data: dict[str, typing.Any],
        serialize: bool = True,
    ) -> dict[str, typing.Any]:
        return {
            k: self(v, serialize=serialize) for k, v in data.items() if type(v) not in (NoneType, Nothing)
        }

    def convert_lst(
        self,
        data: list[typing.Any],
        serialize: bool = True,
    ) -> str | list[typing.Any]:
        converted_lst = [self(x, serialize=False) for x in data]
        return encoder.encode(converted_lst) if serialize is True else converted_lst

    def convert_tpl(self, data: tuple[typing.Any, ...], _: bool = False) -> str | tuple[typing.Any, ...]:
        match data:
            case (str(filename), bytes(content)):
                attach_name = secrets.token_urlsafe(16)
                self._files[attach_name] = (filename, content)
                return "attach://{}".format(attach_name)

        return data


class Proxy:
    def __init__(self, cfg: "_ProxiedDict", key: str) -> None:
        self.key = key
        self.cfg = cfg

    def get(self) -> typing.Any | None:
        return self.cfg._defaults.get(self.key)


class _ProxiedDict(typing.Generic[T]):
    def __init__(self, tp: type[T]) -> None:
        self.type = tp
        self._defaults = {}

    def __setattribute__(self, name: str, value: typing.Any, /) -> None:
        self._defaults[name] = value

    def __getitem__(self, key: str, /) -> None:
        return Proxy(self, key)  # type: ignore

    def __setitem__(self, key: str, value: typing.Any, /) -> None:
        self._defaults[key] = value


if typing.TYPE_CHECKING:

    def ProxiedDict(typed_dct: type[T]) -> T | _ProxiedDict[T]:  # noqa: N802
        ...

else:
    ProxiedDict = _ProxiedDict


__all__ = (
    "DataConverter",
    "MODEL_CONFIG",
    "Model",
    "ProxiedDict",
    "Proxy",
    "full_result",
    "generate_random_id",
    "get_params",
)



import typing

from fntypes.result import Error, Ok, Result

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter.abc import ABCAdapter
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.bot.rules.adapter.raw_update import RawUpdateAdapter
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Model, Update

ToCute = typing.TypeVar("ToCute", bound=BaseCute)


class EventAdapter(ABCAdapter[Update, ToCute]):
    ADAPTED_VALUE_KEY: str = "_adapted_cute_event"

    def __init__(self, event: UpdateType | type[Model], cute_model: type[ToCute]) -> None:
        self.event = event
        self.cute_model = cute_model

    def __repr__(self) -> str:
        raw_update_type = (
            f"Update -> {self.event.__name__}"
            if isinstance(self.event, type)
            else f"Update.{self.event.value}"
        )
        return "<{}: adapt {} -> {}>".format(
            self.__class__.__name__,
            raw_update_type,
            self.cute_model.__name__,
        )

    def get_event(self, update: UpdateCute) -> Model | None:
        if isinstance(self.event, UpdateType) and self.event == update.update_type:
            return update.incoming_update

        if not isinstance(self.event, UpdateType) and (event := update.get_event(self.event)):
            return event.unwrap()

        return None

    def adapt(self, api: API, update: Update, context: Context) -> Result[ToCute, AdapterError]:
        match RawUpdateAdapter().adapt(api, update, context):
            case Ok(update_cute) if event := self.get_event(update_cute):
                if self.ADAPTED_VALUE_KEY in context:
                    return Ok(context[self.ADAPTED_VALUE_KEY])

                adapted = (
                    typing.cast(ToCute, event)
                    if isinstance(event, BaseCute)
                    else self.cute_model.from_update(event, bound_api=api)
                )
                context[self.ADAPTED_VALUE_KEY] = adapted
                return Ok(adapted)
            case Error(_) as err:
                return err
            case _:
                return Error(AdapterError(f"Update is not an {self.event!r}."))


__all__ = ("EventAdapter",)



import typing

from fntypes.result import Error, Ok

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.msgspec_utils import msgspec_convert
from telegrinder.node.base import ComposeError, FactoryNode, Name, ScalarNode
from telegrinder.node.update import UpdateNode

FieldType = typing.TypeVar("FieldType")


class CallbackQueryNode(ScalarNode, CallbackQueryCute):
    @classmethod
    def compose(cls, update: UpdateNode) -> CallbackQueryCute:
        if not update.callback_query:
            raise ComposeError("Update is not a callback_query.")
        return update.callback_query.unwrap()


class CallbackQueryData(ScalarNode, dict[str, typing.Any]):
    @classmethod
    def compose(cls, callback_query: CallbackQueryNode) -> dict[str, typing.Any]:
        return callback_query.decode_callback_data().expect(
            ComposeError("Cannot complete decode callback query data.")
        )


class _Field(FactoryNode):
    field_type: type[typing.Any]

    def __class_getitem__(cls, field_type: type[typing.Any], /) -> typing.Self:
        return cls(field_type=field_type)

    @classmethod
    def compose(cls, callback_query_data: CallbackQueryData, data_name: Name) -> typing.Any:
        if data := callback_query_data.get(data_name):
            match msgspec_convert(data, cls.field_type):
                case Ok(value):
                    return value
                case Error(err):
                    raise ComposeError(err)

        raise ComposeError(f"Cannot find callback data with name {data_name!r}.")


if typing.TYPE_CHECKING:
    Field = typing.Annotated[FieldType, ...]
else:
    Field = _Field


__all__ = ("CallbackQueryData", "CallbackQueryNode", "Field")



import typing

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.dispatch.waiter_machine.hasher.hasher import Hasher
from telegrinder.bot.scenario.checkbox import Checkbox, ChoiceCode

if typing.TYPE_CHECKING:
    from telegrinder.api.api import API
    from telegrinder.bot.dispatch.view.base import BaseStateView
    from telegrinder.bot.scenario.checkbox import Key

    class Choice(Checkbox[Key], typing.Generic[Key]):
        async def wait(
            self,
            hasher: Hasher[CallbackQueryCute, int],
            api: API,
            view: BaseStateView[CallbackQueryCute],
        ) -> tuple[Key, int]: ...

else:

    class Choice(Checkbox):
        async def handle(self, cb):
            code = cb.data.unwrap().replace(self.random_code + "/", "", 1)
            if code == ChoiceCode.READY:
                return False

            for choice in self.choices:
                choice.is_picked = False

            for i, choice in enumerate(self.choices):
                if choice.code == code:
                    self.choices[i].is_picked = True
                    await cb.ctx_api.edit_message_text(
                        text=self.message,
                        chat_id=cb.message.unwrap().v.chat.id,
                        message_id=cb.message.unwrap().v.message_id,
                        parse_mode=self.PARSE_MODE,
                        reply_markup=self.get_markup(),
                    )

            return True

        async def wait(self, hasher, api, view):
            if len(tuple(choice for choice in self.choices if choice.is_picked)) != 1:
                raise ValueError("Exactly one choice must be picked")
            choices, m_id = await super().wait(hasher, api, view)
            return tuple(choices.keys())[tuple(choices.values()).index(True)], m_id


__all__ = ("Choice",)
